#!/usr/bin/perl

#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
no strict 'refs';
use threads;
use threads::shared;

use constant { # Need them up here, before anybody derefs them.
	ST_PRECONNECT => 0,
	ST_LOADMOD => 1,
	ST_NORMAL => 2,
	ST_SHUTDOWN => 3,
	ST_CLOSED => 4,

	DEBUG => 0,
};

use Cwd 'abs_path';
use File::Basename;

use libs::config;
use libs::thread;
use libs::event;
use libs::net;
use libs::help;
use libs::modes;
use libs::log;
use libs::email;
use libs::fmt;
use libs::misc;
use libs::module;

use IO::Handle;

our $status : shared = ST_PRECONNECT;

our $prefix = dirname(abs_path($0));
*notice = \&nickserv::notice;

config::loadconfig();

STDOUT->autoflush(1);

our $progname = 'SurrealServices';
our $version = '0.3.13';
our $extraversion = 'configured for UnrealIRCd 3.2.5 and up';
our $rsnick = 'ServServ';

#module::load();

print "Starting $progname $version.\n";

unless(DEBUG) {
	exit if fork;
	close STDIN;
	close STDOUT;
}

my $spawner; my $writer;

$spawner = threads->new(\&thread::spawn);

net::agent_connect($rsnick, 'service', undef, '+ABHSNaopqz', 'Services Control Agent');

event::addhandler('SEOS', undef, undef, __PACKAGE__, 'main::ev_connect');

sub ev_connect {
	net::join($rsnick, $config::conf{diag});
	net::setmode($rsnick, $config::conf{diag}, '+o', $rsnick);
	event::addtimer('perlserv__pingtimer', 60, __PACKAGE__,
			"main::pingtimer");
}

event::addhandler('PRIVMSG', undef, 'servserv', __PACKAGE__, 'main::dispatch');

sub dispatch {
	my ($src, $dst, $msg) = @_;
	my $user = { NICK => $src, AGENT => $rsnick };
	if(!adminserv::is_ircop($user)) {
		notice($user, 'Access Denied');
		net::globops($rsnick, "\002$src\002 failed access to $rsnick $msg");
		return;
	}
	if($msg =~ /^lsmod/i) {
		notice($user, $config::conf{load});
	}

	if($msg =~ /^lsjoin/i) {
		foreach my $agent (keys(%net::agents)) {
			notice($user, "$agent: " . join(', ', keys(%{$net::agents{lc $agent}{CHANS}})));
		}
	}

	if($msg =~ /^shutdown/i) {
		if(!adminserv::is_svsop($user, adminserv::S_ADMIN() )) {
			notice($user, 'You do not have sufficient rank for this command');
			return;
		}
		
		thread::shutdown();
	}
	if($msg =~ /^raw (.*)/i) {
		if(!adminserv::is_svsop($user, adminserv::S_ROOT() )) {
			notice($user, 'You do not have sufficient rank for this command');
			return;
		}
		net::ircsend($1);
	}
}

#net::serv_connect();

sub pingtimer($) {
	net::ping();
	event::addtimer('perlserv__pingtimer', 60, __PACKAGE__, 
			"main::pingtimer");
}

thread::wait_for_worker();

print "Connecting...";

net::serv_connect();
print " done.\n";

$writer = threads->new(\&thread::writer);

#$timer = threads->new(\&event::checktimers);
#event::checktimers();
#net::serv_connect();

#print " done.\n\n";

net::ircrecv();

$spawner->join;
$writer->join;
